import os
import sys
import socket
import ssl
import errno
import network_serializer

def parse_req(req):
	return network_serializer.loads(req)

def format_req(method, kwargs):
	return network_serializer.dumps((method, kwargs))

def parse_resp(resp):
	return network_serializer.loads(resp)

def format_resp(resp):
	return network_serializer.dumps(resp)

def buffered_readlines(sock):
	buf = ''
	while True:
		while '\n' in buf:
			(line, nl, buf) = buf.partition('\n')
			yield line
		try:
			newdata = sock.recv(4096)
			if newdata == '':
				break
			buf += newdata
		except IOError, e:
			if e.errno == errno.ECONNRESET:
				break

class RpcServer(object):
	def run_sock(self, sock):
		lines = buffered_readlines(sock)
		for req in lines:
			(method, kwargs) = parse_req(req)
			m = self.__getattribute__('rpc_' + method)
			ret = m(**kwargs)
			sock.sendall(format_resp(ret) + '\n')

	def run_socket_fork(self, port):
		server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		server.bind(('', port))

		server.listen(5)
		while True:
			conn, addr = server.accept()
			ssl_conn = ssl.wrap_socket(conn,
						   server_side=True,
						   certfile='server_tls_cert',
						   keyfile='server_tls_key',
						   ssl_version=ssl.PROTOCOL_TLSv1)
			self.run_sock(ssl_conn)
			ssl_conn.shutdown(socket.SHUT_RDWR)
			ssl_conn.close()
		## The server isn't stateless, so forking is sad without more effort
		#	 pid = os.fork()
		#	 if pid == 0:
		#		 # fork again to avoid zombies
		#		 if os.fork() <= 0:
		#			 self.run_sock(conn)
		#			 sys.exit(0)
		#		 else:
		#			 sys.exit(0)
		#	 conn.close()
		#	 os.waitpid(pid, 0)

class RpcClient(object):
	def __init__(self, sock):
		self.sock = sock
		self.lines = buffered_readlines(sock)

	def call(self, method, **kwargs):
		self.sock.sendall(format_req(method, kwargs) + '\n')
		return parse_resp(self.lines.next())

	def close(self):
		self.sock.close()

	## __enter__ and __exit__ make it possible to use RpcClient()
	## in a "with" statement, so that it's automatically closed.
	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.close()

def client_connect(host, port):
	sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	ssl_sock = ssl.wrap_socket(sock)
	ssl_sock.connect((host, port))
	return RpcClient(ssl_sock)

